<?php
/* Copyright (C) 2022 Ravi Trébuchet <ravi@code42.fr>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <https://www.gnu.org/licenses/>.
 */

/**
 * \file        class/fastproduct.class.php
 * \ingroup     produitrapide
 * \brief       This file is a CRUD class file for FastProduct (Create/Read/Update/Delete)
 */

// Put here all includes required by your class file
require_once DOL_DOCUMENT_ROOT . '/product/class/product.class.php';
require_once DOL_DOCUMENT_ROOT . '/core/lib/files.lib.php';
dol_include_once('/produitrapide/lib/produitrapide_fastproduct.lib.php');

/**
 * Class for FastProduct
 */
class FastProduct extends Product
{
	/**
	 * @var string ID of module.
	 */
	public $module = 'produitrapide';

	/**
	 * @var int  Does this object support multicompany module ?
	 * 0=No test on entity, 1=Test with field entity, 'field@table'=Test with link by field@table
	 */
	public $ismultientitymanaged = 0;

	/**
	 * @var string String with name of icon for fastproduct. Must be a 'fa-xxx' fontawesome code (or 'fa-xxx_fa_color_size') or 'fastproduct@produitrapide' if picto is file 'img/object_fastproduct.png'.
	 */
	public $picto = 'fa-file';

	/**
	 *  'type' field format:
	 *    'integer', 'integer:ObjectClass:PathToClass[:AddCreateButtonOrNot[:Filter[:Sortfield]]]',
	 *    'select' (list of values are in 'options'),
	 *    'sellist:TableName:LabelFieldName[:KeyFieldName[:KeyFieldParent[:Filter[:Sortfield]]]]',
	 *    'chkbxlst:...',
	 *    'varchar(x)',
	 *    'text', 'text:none', 'html',
	 *    'double(24,8)', 'real', 'price',
	 *    'date', 'datetime', 'timestamp', 'duration',
	 *    'boolean', 'checkbox', 'radio', 'array',
	 *    'mail', 'phone', 'url', 'password', 'ip'
	 *        Note: Filter can be a string like "(t.ref:like:'SO-%') or (t.date_creation:<:'20160101') or (t.nature:is:NULL)"
	 *  'label' the translation key.
	 *  'picto' is code of a picto to show before value in forms
	 *  'enabled' is a condition when the field must be managed (Example: 1 or '$conf->global->MY_SETUP_PARAM' or '!empty($conf->multicurrency->enabled)' ...)
	 *  'position' is the sort order of field.
	 *  'notnull' is set to 1 if not null in database. Set to -1 if we must set data to null if empty ('' or 0).
	 *  'visible' says if field is visible in list (Examples: 0=Not visible, 1=Visible on list and create/update/view forms, 2=Visible on list only, 3=Visible on create/update/view form only (not list), 4=Visible on list and update/view form only (not create). 5=Visible on list and view only (not create/not update). Using a negative value means field is not shown by default on list but can be selected for viewing)
	 *  'noteditable' says if field is not editable (1 or 0)
	 *  'default' is a default value for creation (can still be overwrote by the Setup of Default Values if field is editable in creation form). Note: If default is set to '(PROV)' and field is 'ref', the default value will be set to '(PROVid)' where id is rowid when a new record is created.
	 *  'index' if we want an index in database.
	 *  'foreignkey'=>'tablename.field' if the field is a foreign key (it is recommanded to name the field fk_...).
	 *  'searchall' is 1 if we want to search in this field when making a search from the quick search button.
	 *  'isameasure' must be set to 1 or 2 if field can be used for measure. Field type must be summable like integer or double(24,8). Use 1 in most cases, or 2 if you don't want to see the column total into list (for example for percentage)
	 *  'css' and 'cssview' and 'csslist' is the CSS style to use on field. 'css' is used in creation and update. 'cssview' is used in view mode. 'csslist' is used for columns in lists. For example: 'css'=>'minwidth300 maxwidth500 widthcentpercentminusx', 'cssview'=>'wordbreak', 'csslist'=>'tdoverflowmax200'
	 *  'help' is a 'TranslationString' to use to show a tooltip on field. You can also use 'TranslationString:keyfortooltiponlick' for a tooltip on click.
	 *  'showoncombobox' if value of the field must be visible into the label of the combobox that list record
	 *  'disabled' is 1 if we want to have the field locked by a 'disabled' attribute. In most cases, this is never set into the definition of $fields into class, but is set dynamically by some part of code.
	 *  'arrayofkeyval' to set a list of values if type is a list of predefined values. For example: array("0"=>"Draft","1"=>"Active","-1"=>"Cancel"). Note that type can be 'integer' or 'varchar'
	 *  'autofocusoncreate' to have field having the focus on a create form. Only 1 field should have this property set to 1.
	 *  'comment' is not used. You can store here any text of your choice. It is not used by application.
	 *    'validate' is 1 if need to validate with $this->validateField()
	 *  'copytoclipboard' is 1 or 2 to allow to add a picto to copy value into clipboard (1=picto after label, 2=picto after value)
	 *
	 *  Note: To have value dynamic, you can set value to 0 in definition and edit the value on the fly into the constructor.
	 */

	// BEGIN MODULEBUILDER PROPERTIES
	/**
	 * @var array  Array with all fields and their property. Do not use it as a static var. It may be modified by constructor.
	 */
	public $fields = array(
		'rowid' => array('type' => 'integer', 'label' => 'TechnicalID', 'enabled' => 1, 'visible' => -2, 'notnull' => 1, 'index' => 1, 'position' => 1, 'comment' => 'Id'),
		'fp_type' => array('type' => 'integer', 'label' => 'ProduitRapideObjectType', 'enabled' => '!empty($conf->produit->enabled) && !empty($conf->service->enabled)', 'visible' => 1, 'notnull' => 1, 'position' => 5),
		'fp_ref' => array('type' => 'varchar(128)', 'label' => 'Ref', 'enabled' => 1, 'visible' => 1, 'notnull' => 1, 'showoncombobox' => 1, 'index' => 1, 'position' => 10, 'searchall' => 1, 'comment' => 'Reference of object', 'placeholder' => ''),
		'entity' => array('type' => 'integer', 'label' => 'Entity', 'enabled' => 1, 'visible' => 0, 'default' => 1, 'notnull' => 1, 'index' => 1, 'position' => 5),
		'fp_label' => array('type' => 'varchar(255)', 'label' => 'Label', 'enabled' => 1, 'visible' => 1, 'notnull' => 1, 'showoncombobox' => 2, 'position' => 15, 'placeholder' => ''),
		'fp_description' => array('type' => 'varchar(255)', 'label' => 'Description', 'enabled' => '$conf->global->PRODUITRAPIDE_DISPLAY_DESCRIPTION', 'visible' => 1, 'notnull' => 0, 'showoncombobox' => 2, 'position' => 20),
		'fp_status' => array('type' => 'integer', 'label' => 'ProduitRapideStatusSell', 'enabled' => 1, 'visible' => 1, 'notnull' => 1, 'default' => 1, 'index' => 1, 'position' => 1000, 'arrayofkeyval' => array(1 => 'ProduitRapideOnSell', 0 => 'ProduitRapideNotOnSell')),
		'fp_status_buy' => array('type' => 'integer', 'label' => 'ProduitRapideStatusBuy', 'enabled' => 1, 'visible' => 1, 'notnull' => 1, 'default' => 1, 'index' => 1, 'position' => 1000, 'arrayofkeyval' => array(1 => 'ProduitRapideOnBuy', 0 => 'ProduitRapideNotOnBuy')),
		'fp_url' => array('type' => 'url', 'label' => 'Url', 'enabled' => '$conf->global->PRODUITRAPIDE_DISPLAY_URL', 'visible' => 1, 'notnull' => 0, 'showoncombobox' => 2, 'position' => 21, 'placeholder' => ''),
		'fp_price' => array('type' => 'price', 'label' => 'ProduitRapideSellPrice', 'enabled' => '$conf->global->PRODUITRAPIDE_DISPLAY_PRICE', 'position' => 23, 'visible' => 1, 'showoncombobox' => 3, 'placeholder' => ''),
		'fp_fk_supplier' => array('type' => 'integer:Societe:societe/class/societe.class.php:1:fournisseur=1', 'label' => 'Supplier', 'enabled' => '!empty($conf->fournisseur->enabled) && $conf->global->PRODUITRAPIDE_DISPLAY_SUPPLIER', 'position' => 24, 'visible' => 1, 'showoncombobox' => 3),
		'fp_supplierref' => array('type' => 'varchar(255)', 'label' => 'ProduitRapideSupplierReference', 'enabled' => '!empty($conf->fournisseur->enabled) && $conf->global->PRODUITRAPIDE_DISPLAY_SUPPLIER', 'position' => 25, 'visible' => 1, 'showoncombobox' => 3, 'placeholder' => ''),
		'fp_supplierprice' => array('type' => 'price', 'label' => 'ProduitRapideSupplierPrice', 'enabled' => '!empty($conf->fournisseur->enabled) && $conf->global->PRODUITRAPIDE_DISPLAY_SUPPLIER', 'position' => 26, 'visible' => 1, 'showoncombobox' => 3, 'placeholder' => ''),
		'fp_files' => array('type' => 'files', 'label' => 'ProduitRapideFiles', 'enabled' => '$conf->global->PRODUITRAPIDE_DISPLAY_FILES', 'notnull' => 0, 'position' => 27, 'visible' => 1, 'showoncombobox' => 3),
		'fp_barcode' => array('type' => 'varchar(255)', 'label' => 'ProduitRapideBarcode', 'enabled' => '!empty($conf->barcode->enabled) && $conf->global->PRODUITRAPIDE_DISPLAY_BARCODE', 'position' => 30, 'visible' => 1, 'showoncombobox' => 3),
		'fp_barcode_type' => array('type' => "sellist:c_barcode_type:code:rowid:0:coder != '0'", 'label' => 'BarcodeType', 'enabled' => '!empty($conf->barcode->enabled) && $conf->global->PRODUITRAPIDE_DISPLAY_BARCODE', 'position' => 29, 'notnull' => 0, 'visible' => -1, 'placeholder' => ''),
		'datec' => array('type' => 'datetime', 'label' => 'DateCreation', 'enabled' => 1, 'visible' => -2, 'notnulxl' => 1, 'position' => 500),
		'tms' => array('type' => 'timestamp', 'label' => 'DateModification', 'enabled' => 1, 'visible' => -2, 'notnull' => 1, 'position' => 501),
		'fk_user_author' => array('type' => 'integer', 'label' => 'UserAuthor', 'enabled' => 1, 'visible' => -2, 'notnull' => 1, 'position' => 510, 'foreignkey' => 'llx_user.rowid'),
	);

	public $rowid;
	public $fk_supplier;
	public $supplierref;
	public $supplierprice;
	public $suppliervat;
	public $suppliervatlocal1;
	public $suppliervatlocal2;
	public $suppliervatlocaltype1;
	public $suppliervatlocaltype2;
	public $suppliervatcode;
	public $supplierpricetype;

	public $fk_propal;
	public $fk_invoice;
	public $fk_order;
	public $fk_order_supplier;
	// END MODULEBUILDER PROPERTIES

	/**
	 * @param  Database    $db    The database
	 */
	public function __construct($db)
	{
		global $langs;

		$this->db = $db;
		$this->canvas = '';

		// Translation of select keys on creation
		foreach ($this->fields as $key => $field) {
			if (isset($field['arrayofkeyval']) && is_array($field['arrayofkeyval'])) {
				foreach ($field['arrayofkeyval'] as $keyval => $val) {
					$this->fields[$key]['arrayofkeyval'][$keyval] = $langs->trans($val);
				}
			}
		}
	}

	/**
	 *    Insert product into database
	 *
	 * @param  User    $user         User making insert
	 * @param  int     $notrigger    Disable triggers
	 * @return int             Returns id of product/service if OK, < 0 if KO
	 */
	public function create($user, $notrigger = 0)
	{
		global $conf;
		if ($this->supplierref && supplierRefExists($this->supplierref)) {
			return -5;
		}
		// Creation of Product (Fast Product is not registered in database)
		$res = Product::create($user, $notrigger);

		// If product creation worked
		if ($res > 0) {
			// If Fast Product as a supplier, we add a supplier price to the created product
			if (isset($this->fk_supplier)) {
				// If supplier price type isn't set, we set it to 'HT'
				if (empty($this->supplierpricetype)) $this->supplierpricetype = 'HT';
				if (!isset($this->suppliervatlocal1)) $this->suppliervatlocal1 = 0;
				if (!isset($this->suppliervatlocal2)) $this->suppliervatlocal2 = 0;
				if (!isset($this->suppliervatlocaltype1)) $this->suppliervatlocaltype1 = 0;
				if (!isset($this->suppliervatlocaltype2)) $this->suppliervatlocaltype2 = 0;
				if (!isset($this->suppliervatcode)) $this->suppliervatcode = '';
				// We add the buy price
				addFastProductSupplier($res, $this->fk_supplier, $this->supplierref, $this->supplierprice, $this->suppliervat, $this->suppliervatlocal1, $this->suppliervatlocal2, $this->suppliervatlocaltype1, $this->suppliervatlocaltype2, $this->suppliervatcode, $this->supplierpricetype);
			}
			// If Category is enabled, we add "Fast Product" tag
			if ($conf->category->enabled ?? false) {
				addFastProductTag($res);
			}
			// If product was created from a propal, we add a link between the new product and the propal
			if (!empty($this->fk_propal)) {
				addFastProductPropal($res, $this->fk_propal);
			}
			// If product was created from an invoice, we add a link between the new product and the invoice
			if (!empty($this->fk_invoice)) {
				addFastProductInvoice($res, $this->fk_invoice);
			}
			// If product was created from an order, we add a link between the new product and the order
			if (!empty($this->fk_order)) {
				addFastProductOrder($res, $this->fk_order);
			}
			// If product was created from a supplier order, we add a link between the new product and the supplier order
			if (!empty($this->fk_order_supplier)) {
				addFastProductOrderSupplier($res, $this->fk_order_supplier);
			}
		}
		return $res;
	}

	/**
	 * Create product from FastProduct
	 *
	 * @param  CommonObject    $object    Invoice or Propal
	 * @return int Product id or -1 if creation failed
	 */
	public function createFastProduct($object)
	{

		global $langs, $conf, $user, $db, $mysoc;

		// Get all product information to create it
		$type = GETPOST('fp_type', 'int');

		$this->type = (($type == 0) ? Product::TYPE_PRODUCT : Product::TYPE_SERVICE);

		// Fields with same name as Product object
		$basicFields = array(
			"ref" => null,
			"label" => null,
			"status" => "int",
			"status_buy" => "int",
			"description" => null,
			"url" => null,
			"barcode" => null,
			"barcode_type" => "int"
		);
		// Loop on basic fields to GET values
		foreach ($basicFields as $field => $check) {
			$this->$field = GETPOST('fp_' . $field, ($check ?? "alphanohtml"));
		}

		$this->price_base_type = GETPOST('fp_price_type', 'alphanohtml');
		$tva_tx_txt = GETPOST('fp_vat_rate', 'alphanohtml');
		$this->tva_tx = preg_replace('/[^0-9\.].*$/', '', $tva_tx_txt); // keep remove all after the numbers and dot
		$this->tva_npr = preg_match('/\*/', $tva_tx_txt) ? 1 : 0;
		$reg = array();
		if (preg_match('/\((.*)\)/', $tva_tx_txt, $reg)) {
			// We look into database using code (we can't use get_localtax() because it depends on buyer that is not known). Same in update price.
			$vatRateCode = $reg[1];
			$obj = getVatRateObject($vatRateCode, $this->tva_tx, $mysoc->country_code);
			if (!empty($obj)) {
				$this->tva_npr = $obj->recuperableonly;
				$this->localtax1_tx = $obj->localtax1;
				$this->localtax2_tx = $obj->localtax2;
				$this->localtax1_type = $obj->localtax1_type;
				$this->localtax2_type = $obj->localtax2_type;
				$this->default_vat_code = $vatRateCode;
			}
		}

		$price = GETPOST('fp_price');
		if ($this->price_base_type == "HT" && $price) {
			$this->price = price2num($price);
			$this->price_ttc = ($this->price + ($this->price * ((int) $this->tva_tx / 100)));
		} else if ($price) {
			$this->price_ttc = price2num($price);
			$this->price = ($this->price_ttc / (1 + ((int) $this->tva_tx / 100)));
		} else {
			$this->price = 0;
		}
		$this->fk_supplier = GETPOST('fp_fk_supplier', 'int');
		$this->supplierref = GETPOST('fp_supplierref', 'alphanohtml');
		$this->supplierprice = GETPOST('fp_supplierprice', 'alphanohtml');
		$tva_tx_txt = GETPOST('fp_vat_rate_supplier', 'alphanohtml');
		$this->suppliervat = preg_replace('/[^0-9\.].*$/', '', $tva_tx_txt); // keep remove all after the numbers and dot
		$reg = array();
		if (preg_match('/\((.*)\)/', $tva_tx_txt, $reg)) {
			// We look into database using code (we can't use get_localtax() because it depends on buyer that is not known). Same in update price.
			$vatRateCode = $reg[1];
			$supplier = new Societe($db);
			$supplier->fetch($this->fk_supplier);
			$obj = getVatRateObject($vatRateCode, $this->suppliervat, $supplier->country_code);
			if (!empty($obj)) {
				$this->tva_npr = $obj->recuperableonly;
				$this->suppliervatlocal1 = $obj->localtax1;
				$this->suppliervatlocal2 = $obj->localtax2;
				$this->suppliervatlocaltype1 = $obj->localtax1_type;
				$this->suppliervatlocaltype2 = $obj->localtax2_type;
				$this->suppliervatcode = $vatRateCode;
			}
		}
		$this->supplierpricetype = GETPOST('fp_supplierprice_type', 'alphanohtml');
		$this->fk_default_warehouse = GETPOST('fp_default_warehouse', 'int');

		// If tva doesn't exist, set it to 0
		if ($this->tva_tx == '') $this->tva_tx = 0;
		if ($this->suppliervat == '') $this->suppliervat = 0;

		// Set default barcode if activated
		if ($conf->barcode->enabled ?? false) {
			if ($this->barcode_type == 0) {
				if (empty($conf->produitrapide->PRODUITRAPIDE_ENABLE_BUTTON_PROPAL) || !$conf->produitrapide->PRODUITRAPIDE_ENABLE_BUTTON_PROPAL) {
					$this->barcode_type = getDefaultBarcodeType();
				} else {
					$this->barcode_type = null;
				}
			}
			if ($this->barcode == "" && (empty($conf->produitrapide->PRODUITRAPIDE_ENABLE_BUTTON_PROPAL) || !$conf->produitrapide->PRODUITRAPIDE_ENABLE_BUTTON_PROPAL)) {
				$this->barcode = getDefaultBarcodeValue();
			}
		}

		// [#20] Add accountancy code
		if (!empty($conf->accounting->enabled)) {
			$this->accountancy_code_buy = (($type == 0) ? getDolGlobalString("ACCOUNTING_PRODUCT_BUY_ACCOUNT") : getDolGlobalString("ACCOUNTING_SERVICE_BUY_ACCOUNT"));
			$this->accountancy_code_buy_intra = (($type == 0) ? getDolGlobalString("ACCOUNTING_PRODUCT_BUY_INTRA_ACCOUNT") : getDolGlobalString("ACCOUNTING_SERVICE_BUY_INTRA_ACCOUNT"));
			$this->accountancy_code_sell = (($type == 0) ? getDolGlobalString("ACCOUNTING_PRODUCT_SOLD_ACCOUNT") : getDolGlobalString("ACCOUNTING_SERVICE_SOLD_ACCOUNT"));
			$this->accountancy_code_sell_intra = (($type == 0) ? getDolGlobalString("ACCOUNTING_PRODUCT_SOLD_INTRA_ACCOUNT") : getDolGlobalString("ACCOUNTING_SERVICE_SOLD_INTRA_ACCOUNT"));
		}

		// If the current object is an invoice or a propal, complete $this information
		if (get_class($object) == "Facture") {
			$this->fk_invoice = $object->id;
		} else if (get_class($object) == "Propal") {
			$this->fk_propal = $object->id;
		} else if (get_class($object) == "Commande") {
			$this->fk_order = $object->id;
		} else if (get_class($object) == "CommandeFournisseur") {
			$this->fk_order_supplier = $object->id;
		}

		// Create the product
		$res = $this->create($user);
		// If product was created
		if ($res > 0) {
			// Check if files were sent
			if (!empty($_FILES) && count($_FILES) > 0 && isset($_FILES['userfile']['tmp_name'][0]) && $_FILES['userfile']['tmp_name'][0] != '') {
				// Complete $file variable
				if (is_array($_FILES['userfile']['tmp_name'])) {
					$files = $_FILES['userfile']['tmp_name'];
				} else {
					$files = array($_FILES['userfile']['tmp_name']);
				}

				// Create a product and fetch the created one to add files to it
				$prod = new Product($db);
				$prod->fetch($res);

				// Foreach file, link it
				foreach ($files as $key => $userfile) {
					if (empty($_FILES['userfile']['tmp_name'][$key])) {
						$error++;
						if ($_FILES['userfile']['error'][$key] == 1 || $_FILES['userfile']['error'][$key] == 2) {
							setEventMessages($langs->trans('ErrorFileSizeTooLarge'), null, 'errors');
						} else {
							setEventMessages($langs->trans("ErrorFieldRequired", $langs->transnoentitiesnoconv("File")), null, 'errors');
						}
					}
					if (preg_match('/__.*__/', $_FILES['userfile']['name'][$key])) {
						$error++;
						setEventMessages($langs->trans('ErrorWrongFileName'), null, 'errors');
					}
				}

				// Set upload
				$upload_dir = $conf->product->multidir_output[$prod->entity] . '/' . get_exdir(0, 0, 0, 1, $prod, 'product');

				// Add file if it was found
				if (empty($error)) {
					if (!empty($upload_dir)) {
						$result = dol_add_file_process($upload_dir, 1, 1, 'userfile', '', null, '', 0, $prod);
					}
				}
				if (isset($result) && $result <= 0) {
					setEventMessages($langs->trans('ProduitRapideFastProductFilesError'), '', 'warnings');
				}
			}
			setEventMessages($langs->trans('ProduitRapideFastProductCreated'), '');
			return $res;
		} else {
			if ($res === -5) {
				setEventMessages($langs->trans('ProduitRapideSupplierRefAlreadyExists'), '', 'errors');
			} else {
				setEventMessages($langs->trans($this->error), '', 'errors');
			}
			return -1;
		}
	}
}
