<?php
/* Copyright (C) 2022 Ravi Trébuchet <ravi@code42.fr>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */

use h2g2\QueryBuilder;

require_once DOL_DOCUMENT_ROOT . '/fourn/class/fournisseur.class.php';
require_once DOL_DOCUMENT_ROOT . '/product/class/product.class.php';
require_once DOL_DOCUMENT_ROOT . '/comm/propal/class/propal.class.php';
require_once DOL_DOCUMENT_ROOT . '/compta/facture/class/facture.class.php';
require_once DOL_DOCUMENT_ROOT . '/commande/class/commande.class.php';
require_once DOL_DOCUMENT_ROOT . '/fourn/class/fournisseur.commande.class.php';
dol_include_once('h2g2/class/querybuilder.class.php');

/**
 * \file    lib/produitrapide_fastproduct.lib.php
 * \ingroup produitrapide
 * \brief   Library files with common functions for FastProduct
 */

/**
 * Prepare array of tabs for FastProduct
 *
 * @param  FastProduct    $object    FastProduct
 * @return    array                    Array of tabs
 */
function fastproductPrepareHead($object)
{
	global $db, $langs, $conf;

	$langs->load("produitrapide@produitrapide");

	$h = 0;
	$head = array();

	$head[$h][0] = dol_buildpath("/produitrapide/fastproduct_card.php", 1) . '?id=' . $object->id;
	$head[$h][1] = $langs->trans("Card");
	$head[$h][2] = 'card';
	$h++;

	complete_head_from_modules($conf, $langs, $object, $head, $h, 'fastproduct@produitrapide');

	complete_head_from_modules($conf, $langs, $object, $head, $h, 'fastproduct@produitrapide', 'remove');

	return $head;
}

/**
 * Gets the id of "Fast Product" tag
 *
 * @return int Returns the id of "Fast Product" tag
 */
function getFastProductTag()
{
	global $conf;

	// Get fast product Tag
	try {
		$tagId = QueryBuilder::table('categorie')
			->select('rowid')
			->where('label', '=', 'Produit Rapide')
			->where('type', '=', '0')
			->where('entity', '=', $conf->entity)
			->get();

		// If tag was found, return it
		if (is_array($tagId) && count($tagId) > 0) {
			return $tagId[0]->rowid;
		}
	} catch (Exception $e) {
		print $e->getMessage();
	}
	return null;
}

/**
 * Add "Fast Product" tag to the given product
 *
 * @param  int    $id    The id of the product
 * @return int Returns 1 if tag was added and -1 if not
 */
function addFastProductTag(int $id): int
{
	global $db;

	$resql = $db->query("INSERT INTO " . MAIN_DB_PREFIX . "categorie_product (fk_categorie, fk_product) VALUES (" . getFastProductTag() . ", " . $id . ")");

	return $resql != null ? 1 : -1;
}

/**
 * Get the society
 *
 * @param  int    $id    The id of the society
 * @return int Returns 1 if society is a supplier and -1 if not
 */
function isSupplier($id)
{
	global $db;

	$supplier = new Fournisseur($db);
	if ($supplier->fetch($id)) {
		return $supplier->fournisseur;
	} else {
		return -1;
	}
}

/**
 * Add a new buy price to the product
 *
 * @param  int       $productId        The id of product
 * @param  int       $supplierId       The id of invoice
 * @param  string    $supplierRef      The id of invoice
 * @param  int       $supplierPrice    The price
 * @param  int       $tvaRate          The VAT rate
 * @param  int       $tvaLT1Tx         The local taxes 1
 * @param  int       $tvaLT2Tx         The local taxes 2
 * @param  int       $tvaLT1Type       The local taxes type 1
 * @param  int       $tvaLT2Type       The local taxes type 2
 * @param  int       $tvaCode          The default VAT code
 * @param  int       $vatType          The VAT type
 * @return    int                            Returns -1 if failed or 1 if add worked
 */
function addFastProductSupplier($productId, $supplierId, $supplierRef, $supplierPrice, $tvaRate, $tvaLT1Tx = 0, $tvaLT2Tx = 0, $tvaLT1Type = 0, $tvaLT2Type = 0, $tvaCode = "", $vatType = 'HT')
{

	global $db, $conf, $user;

	$tva = $tvaRate;

	if (!empty($supplierId) && isSupplier($supplierId) && !empty($productId) && isset($tva) && isset($supplierRef) && !empty($supplierRef)) {
		// If price is empty, we set it to 0
		if (empty($supplierPrice)) {
			$supplierPrice = 0;
		}
		$supplierPrice = price2num($supplierPrice);
		// If TVA type is "TTC", we must calculate again the price
		if ($vatType == 'TTC') {
			$supplierPrice = round(($supplierPrice / (1 + ((int) $tva / 100))), 2);
		}
		// Buy price add
		$resql = $db->query("INSERT INTO " . MAIN_DB_PREFIX . "product_fournisseur_price (entity, datec, fk_product, fk_soc, ref_fourn, price, unitprice, tva_tx, default_vat_code, localtax1_tx, localtax2_tx, localtax1_type, localtax2_type, fk_user, quantity) VALUES (" . $conf->entity . ", '" . $db->idate(dol_now()) . "', " . $productId . ", " . $supplierId . ", '" . $supplierRef . "', " . $supplierPrice . ", " . $supplierPrice . ", " . $tva . ", '" . $tvaCode . "', " . $tvaLT1Tx . ", " . $tvaLT2Tx . ", " . $tvaLT1Type . ", '" . $tvaLT2Type . "', " . $user->id . ", 1)");
	} else {
		$resql = null;
	}

	return $resql != null ? 1 : -1;
}

/**
 * Link the product to the propal
 *
 * @param  int    $id          The id of product
 * @param  int    $propalId    The id of propal
 * @return int Returns -1 if failed or 1 if add worked
 */
function addFastProductPropal($id, $propalId)
{
	global $db, $conf, $user;


	// Create product
	$product = new Product($db);
	$product->fetch($id);
	// Create propal
	$propal = new Propal($db);
	$propal->fetch($propalId);
	if ($propal->status == Propal::STATUS_VALIDATED) {
		$validate = true;
		$propal->setDraft($user, 1);
	}

	if (isset($product->id) && $product->id > 0) {
		$buyPrice = getFirstBuyPrice($product->id);
		// [#26] Add supplier price on each line
		if ($buyPrice ?? false) {
			$fk_fourn = $buyPrice->rowid;
			$fourn_price = $buyPrice->price;
		} else {
			$fk_fourn = $fourn_price = 0;
		}

		$tvaText = ($product->tva_tx ?? $conf->global->PRODUITRAPIDE_TVA ?? 0);
		$infoBits = 0;

		$obj = getVatRateObject($product->default_vat_code, $tvaText, ($propal->thirdparty->country_code ?? "FR"));
		if (!empty($obj)) {
			$infoBits = $obj->recuperableonly;
			$localTax1 = $obj->localtax1;
			$localTax2 = $obj->localtax2;
			$tvaText .= ($product->default_vat_code ? " (" . $product->default_vat_code . ")" : "");
		}
		// Add product to propal
		$res = $propal->addline($product->description, $product->price, 1, $tvaText, $localTax1 ?? 0.0, $localTax2 ?? 0.0, $product->id, 0, $product->price_base_type, $product->price_ttc, $infoBits, $product->type, -1, 0, 0, $fk_fourn, $fourn_price);
	} else {
		$res = -1;
	}
	if (!empty($validate)) {
		$propal->setStatut(Propal::STATUS_VALIDATED);
	}
	return ($res > 0) ? 1 : -1;
}

/**
 * Link the product to the invoice
 *
 * @param  int    $id           The id of product
 * @param  int    $invoiceId    The id of invoice
 * @return int Returns -1 if failed or 1 if add worked
 */
function addFastProductInvoice($id, $invoiceId)
{
	global $db, $conf, $user;

	// Create product
	$product = new Product($db);
	$product->fetch($id);
	// Create invoice
	$invoice = new Facture($db);
	$invoice->fetch($invoiceId);
	$invoice->fetch_thirdparty();

	if ($invoice->status == Facture::STATUS_VALIDATED) {
		$validate = true;
		$invoice->setDraft($user, 1);
	}

	if (isset($product->id) && $product->id > 0) {
		$buyPrice = getFirstBuyPrice($product->id);
		// [#26] Add supplier price on each line
		if ($buyPrice) {
			$fk_fourn = $buyPrice->rowid;
			$fourn_price = $buyPrice->price;
		} else {
			$fk_fourn = $fourn_price = 0;
		}
		$tvaText = ($product->tva_tx ?? $conf->global->PRODUITRAPIDE_TVA ?? 0);
		$infoBits = 0;

		$obj = getVatRateObject($product->default_vat_code, $tvaText, $invoice->thirdparty->country_code);
		if (!empty($obj)) {
			$infoBits = $obj->recuperableonly;
			$localTax1 = $obj->localtax1;
			$localTax2 = $obj->localtax2;
			$tvaText .= ($product->default_vat_code ? " (" . $product->default_vat_code . ")" : "");
		}
		// Add product to invoice
		$res = $invoice->addline($product->description, $product->price, 1, $tvaText, $localTax1, $localTax2, $product->id, 0, '', '', 0, $infoBits, '', $product->price_base_type, $product->price_ttc, $product->type, -1, 0, '', 0, 0, $fk_fourn, $fourn_price);
	} else {
		$res = -1;
	}
	if ($validate ?? false) {
		$invoice->setStatut(Facture::STATUS_VALIDATED);
	}

	return ($res > 0) ? 1 : -1;
}

/**
 * Link the product to the order
 *
 * @param  int    $id         The id of product
 * @param  int    $orderId    The id of order
 * @return int Returns -1 if failed or 1 if add worked
 */
function addFastProductOrder($id, $orderId)
{
	global $db, $conf, $user;

	// Create product
	$product = new Product($db);
	$product->fetch($id);
	// Create invoice
	$order = new Commande($db);
	$order->fetch($orderId);
	$order->fetch_thirdparty();

	if ($order->status == Commande::STATUS_VALIDATED) {
		$validate = true;
		$order->setDraft($user, 1);
	}

	if (isset($product->id) && $product->id > 0) {
		$buyPrice = getFirstBuyPrice($product->id);
		// [#26] Add supplier price on each line
		if ($buyPrice ?? false) {
			$fk_fourn = $buyPrice->rowid;
			$fourn_price = $buyPrice->price;
		} else {
			$fk_fourn = $fourn_price = 0;
		}
		$tvaText = ($product->tva_tx ?? $conf->global->PRODUITRAPIDE_TVA ?? 0);
		$infoBits = 0;

		$obj = getVatRateObject($product->default_vat_code, $tvaText, $order->thirdparty->country_code);
		if (!empty($obj)) {
			$infoBits = $obj->recuperableonly;
			$localTax1 = $obj->localtax1;
			$localTax2 = $obj->localtax2;
			$tvaText .= ($product->default_vat_code ? " (" . $product->default_vat_code . ")" : "");
		}
		// Add product to invoice
		$res = $order->addline($product->description, $product->price, 1, $tvaText, $localTax1, $localTax2, $product->id, 0, $infoBits, 0, $product->price_base_type, $product->price_ttc, '', '', $product->type, -1, 0, 0, $fk_fourn, $fourn_price);
	} else {
		$res = -1;
	}
	if ($validate ?? false) {
		$order->setStatut(Commande::STATUS_VALIDATED);
	}
	return ($res > 0) ? 1 : -1;
}

/**
 * Link the product to the order
 *
 * @param  int    $id                 The id of product
 * @param  int    $orderSupplierId    The id of supplier order
 * @return int Returns -1 if failed or 1 if add worked
 */
function addFastProductOrderSupplier($id, $orderSupplierId)
{
	global $db, $conf, $user;

	// Create product
	$product = new Product($db);
	$product->fetch($id);
	// Create invoice
	$supplierOrder = new CommandeFournisseur($db);
	$supplierOrder->fetch($orderSupplierId);
	$supplierOrder->fetch_thirdparty();
	if ($supplierOrder->status == CommandeFournisseur::STATUS_VALIDATED) {
		$validate = true;
		$supplierOrder->setDraft($user, 1);
	}
	if (isset($product->id) && $product->id > 0) {
		$buyPrice = getFirstBuyPrice($product->id);
		// [#26] Add supplier price and not selling price
		if ($buyPrice ?? false) {
			$fk_fourn = $buyPrice->rowid;
			$ref_fourn = $buyPrice->ref_fourn;
			$fourn_price = $buyPrice->price;
			$fourn_tva = $buyPrice->tva_tx;
		} else {
			$fk_fourn = $fourn_price = $fourn_tva = 0;
			$ref_fourn = '';
		}
		$tvaText = $fourn_tva;
		$infoBits = 0;

		$obj = getVatRateObject($product->default_vat_code, $tvaText, $supplierOrder->thirdparty->country_code);
		if (!empty($obj)) {
			$infoBits = $obj->recuperableonly;
			$localTax1 = $obj->localtax1;
			$localTax2 = $obj->localtax2;
			$tvaText .= ($product->default_vat_code ? " (" . $product->default_vat_code . ")" : "");
		}
		// Add product to invoice
		$res = $supplierOrder->addline($product->description, $fourn_price, 1, $tvaText, $localTax1, $localTax2, $product->id, $fk_fourn, $ref_fourn, 0, $product->price_base_type, $product->price_ttc, $product->type, $infoBits);
	} else {
		$res = -1;
	}
	if ($validate ?? false) {
		$supplierOrder->setStatut(CommandeFournisseur::STATUS_VALIDATED);
	}

	return ($res > 0) ? 1 : -1;
}

/**
 * Checks if supplier reference already exists
 *
 * @param  string    $ref    The supplier reference to test
 * @return bool True if OK / False if KO
 */
function supplierRefExists($ref)
{
	global $conf;

	// Get fast product Tag
	try {
		$priceRef = QueryBuilder::table('product_fournisseur_price')
			->select('rowid')
			->where('ref_fourn', '=', $ref)
			->where('entity', '=', $conf->entity)
			->get();

		// If tag was found, return it
		if (is_array($priceRef) && count($priceRef) > 0) {
			return true;
		}
	} catch (Exception $e) {
		print $e->getMessage();
	}
	return false;
}

/**
 * Gets first found buy price for the given product
 *
 * @param  int    $prod_id    The id of product
 * @return StdClass Object if price found, Null if not
 */
function getFirstBuyPrice($prod_id)
{
	global $conf;

	// Get fast product Tag
	try {
		$priceId = QueryBuilder::table('product_fournisseur_price')
			->select('rowid', 'price', 'ref_fourn', 'tva_tx')
			->where('fk_product', '=', $prod_id)
			->where('entity', '=', $conf->entity)
			->get();

		// If tag was found, return it
		if (is_array($priceId) && count($priceId) > 0) {
			return $priceId[0];
		}
	} catch (Exception $e) {
		print $e->getMessage();
	}
	return null;
}

/**
 * Get vat rate object to have details on choosen vat rate
 *
 * @param  string    $vatRateCode    Vat rate code
 * @param  float     $vateRate       Vat Rate
 * @param  string    $countryCode    Country code
 * @return Object|null
 */
function getVatRateObject($vatRateCode, $vateRate, $countryCode)
{
	global $db;

	// Get record from code
	$sql = "SELECT t.rowid, t.code, t.recuperableonly, t.localtax1, t.localtax2, t.localtax1_type, t.localtax2_type";
	$sql .= " FROM " . MAIN_DB_PREFIX . "c_tva as t, " . MAIN_DB_PREFIX . "c_country as c";
	$sql .= " WHERE t.fk_pays = c.rowid AND c.code = '" . $db->escape($countryCode) . "'";
	$sql .= " AND t.taux = " . ((float) $vateRate) . " AND t.active = 1";
	$sql .= " AND t.code = '" . $db->escape($vatRateCode) . "'";
	$resql = $db->query($sql);
	if ($resql) {
		$obj = $db->fetch_object($resql);
		return $obj;
	}
	return null;
}
